"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
var _DeQueue = _interopRequireDefault(require("../DeQueue"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
class PQueue {
  constructor(opts) {
    opts = Object.assign({
      concurrency: Infinity,
      autoStart: true,
      queueClass: _DeQueue.default
    }, opts);
    if (!(typeof opts.concurrency === 'number' && opts.concurrency >= 1)) {
      throw new TypeError(`Expected \`concurrency\` to be a number from 1 and up, got \`${opts.concurrency}\` (${typeof opts.concurrency})`);
    }
    this.queue = new opts.queueClass(); // eslint-disable-line new-cap
    this._queueClass = opts.queueClass;
    this._pendingCount = 0;
    this._concurrency = opts.concurrency;
    this._isPaused = opts.autoStart === false;
    this._resolveEmpty = () => {};
    this._resolveIdle = () => {};
  }
  _next() {
    this._pendingCount--;
    if (this.queue.size > 0) {
      if (!this._isPaused) {
        this.queue.dequeue()();
      }
    } else {
      this._resolveEmpty();
      this._resolveEmpty = () => {};
      if (this._pendingCount === 0) {
        this._resolveIdle();
        this._resolveIdle = () => {};
      }
    }
  }
  add(fn, opts) {
    return new Promise((resolve, reject) => {
      const run = () => {
        this._pendingCount++;
        try {
          Promise.resolve(fn()).then(val => {
            resolve(val);
            this._next();
          }, err => {
            reject(err);
            this._next();
          });
        } catch (err) {
          reject(err);
          this._next();
        }
      };
      if (!this._isPaused && this._pendingCount < this._concurrency) {
        run();
      } else {
        this.queue.enqueue(run, opts);
      }
    });
  }
  addAll(fns, opts) {
    return Promise.all(fns.map(fn => this.add(fn, opts)));
  }
  start() {
    if (!this._isPaused) {
      return;
    }
    this._isPaused = false;
    while (this.queue.size > 0 && this._pendingCount < this._concurrency) {
      this.queue.dequeue()();
    }
  }
  pause() {
    this._isPaused = true;
  }
  clear() {
    this.queue = new this._queueClass(); // eslint-disable-line new-cap
  }
  waitEmpty() {
    // Instantly resolve if the queue is empty
    if (this.queue.size === 0) {
      return Promise.resolve();
    }
    return new Promise(resolve => {
      const existingResolve = this._resolveEmpty;
      this._resolveEmpty = () => {
        existingResolve();
        resolve();
      };
    });
  }
  waitIdle() {
    // Instantly resolve if none pending
    if (this._pendingCount === 0) {
      return Promise.resolve();
    }
    return new Promise(resolve => {
      const existingResolve = this._resolveIdle;
      this._resolveIdle = () => {
        existingResolve();
        resolve();
      };
    });
  }
  get size() {
    return this.queue.size;
  }
  get pending() {
    return this._pendingCount;
  }
  get isPaused() {
    return this._isPaused;
  }
}
exports.default = PQueue;
//# sourceMappingURL=queue.js.map